/*----------------------------------------------------------------------------------
----------------------------------------------------------------------------------
// CanadaFsaMap:

 © 2011 - 2015 DOMO, INC.
----------------------------------------------------------------------------------
----------------------------------------------------------------------------------*/
const d3 = require('d3');
const BaseWidget = require('@domoinc/base-widget');
const SummaryNumber = require('@domoinc/summary-number');
const service = new SummaryNumber();
const daTheme = require('@domoinc/da-theme2');
const DomoTooltip = require('@domoinc/domo-tooltip');

const ValueNode = require('./ValueNode');
const GeometryNode = require('./GeometryNode');
const LazyLoad = require('./LazyLoad');
const fsaToProvince = require('./fsaToProvince');
const button = require('./button/button');

const plusSvg = require('raw!./svg/plus-16.svg');
const minusSvg = require('raw!./svg/minus-16.svg');

module.exports = BaseWidget.extend('CanadaFsaMap', {
  //**********************************************************************************
  // Initialization
  //**********************************************************************************
  initialize: function () {
    'use strict';

    const _Chart = this;
    const colorChoices = [
      {name: 'Orange', value: ["#FCCF84", "#FBAD56", "#FB8D34", "#E45621", "#A43724"]},
      {name: 'Blue', value: ["#B7DAF5", "#90c4e4", "#73B0D7", "#4E8CBA", "#31689B"]},
      {name: 'Green', value: ["#BBE491", "#A0D771", "#80C25D", "#559E38", "#387B26"]},
      {name: 'Pink',  value: ['#FBB6DD', '#F395CD', '#EE76BF', '#CF51AC', '#A62A92']},
      {name: 'Average', value: ["#336088", "#71A7C7", "#B7D7E6", "#D8EBF3", "#FFFFFF", "#F7DAD7", "#EFB7B1", "#E37872", "#B53230"]}
    ];
    //----------------------------------------------------------------------------------
    // Config
    //----------------------------------------------------------------------------------
    _Chart._newConfig = {
      //App builder configurable configs
      width: {
        name: 'Widget Width',
        description: 'Width of the widget',
        value: 700,
        type: 'number',
        units: 'px'
      },
      height: {
        name: 'Widget Height',
        description: 'Height of the widget',
        value: 400,
        type: 'number',
        units: 'px'
      },
      colorTheme: {
        name:'Color Theme',
        description: 'The color theme used by the elements',
        value: colorChoices[1],
        type: 'select',
        options: colorChoices
      },
      noDataColor: daTheme.themeElements.noDataColor(),
      districtStrokeColor: {
        name: 'Province Border Color',
        description: 'Color for the province border',
        value: '#333333',
        type: 'color'
      },
      districtStrokeWidth: {
        name: 'Province Border Width',
        description: 'Width of the province border',
        type: 'number',
        value: 1,
        units: 'px'
      },
      digitStrokeColor: {
        name: 'City Border Color',
        description: 'Color for the border around groups of FSAs',
        value: '#585858',
        type: 'color'
      },
      digitStrokeWidth: {
        name: 'City Border Width',
        description: 'Width of the border around groups of FSAs',
        type: 'number',
        value: 1,
        units: 'px'
      },
      regionStrokeColor: {
        name: 'FSA Border Color',
        description: 'Color for the border around individual FSAs',
        value: '#585858',
        type: 'color'
      },
      regionStrokeWidth: {
        name: 'FSA Border Width',
        description: 'Width of the border around indivisual FSAs',
        type: 'number',
        value: 0.5,
        units: 'px'
      },
      highlightStrokeColor: {
        name: 'Highlight Border Color',
        description: 'Color for the border around hovered regions',
        value: '#FDECAD',
        type: 'color'
      },
      highlightStrokeWidth: {
        name: 'Highlight Border Width',
        description: 'Width of the border around hovered regions',
        type: 'number',
        value: 2,
        units: 'px'
      },
      showFsaLabels: {
        name: 'Show FSA Labels',
        description: 'Show or hide the labels on the FSAs',
        type: 'select',
        value: {name: 'Show', value: true},
        options: [
          {name: 'Show', value: true},
          {name: 'Hide', value: false},
        ]
      },
      fsaLabelFontFamily: {
        name: 'FSA Label Font',
        description: 'Font type for the FSA labels',
        category: 'Text',
        value: 'Open Sans',
        type: 'font'
      },
      fsaLabelFontSize: {
        name: 'FSA Label Font Size',
        description: 'Font size for the FSA labels',
        value: 13,
        type: 'number',
        units: 'px'
      },
      fsaLabelFontColor: {
        name: 'FSA Label Font Color',
        description: 'Font color for the FSA labels',
        value: '#333333',
        type: 'color'
      },
      tooltipBackground: daTheme.themeElements.tooltipBackground(),
      tooltipFontFamily: daTheme.themeElements.tooltipFontFamily(),
      tooltipFontSize: daTheme.themeElements.tooltipFontSize(),
      tooltipFontColor: daTheme.themeElements.tooltipFontColor(),
      //Default configs not seen in the app builder
      chartName: {
        description: 'Name of chart for reporting',
        type: 'string',
        value: 'templateChart'
      },
      animationDuration: {
        type: 'number',
        value: 250,
        units: 'ms'
      }
    };
    _Chart.mergeConfig(_Chart._newConfig);

    //----------------------------------------------------------------------------------
    // Data Definition
    //----------------------------------------------------------------------------------
    _Chart._newDataDefinition = {
      'FSA': {
        type: 'string',
        validate: function (d) { return this.accessor(d); },
        accessor: line => line[0]
      },
      'Value': {
        type: 'number',
        validate: function (d) { return (this.accessor(d) || this.accessor(d) === 0); },
        accessor: line => parseFloat(line[1])
      }
    };
    _Chart.mergeDataDefinition(_Chart._newDataDefinition);

    //----------------------------------------------------------------------------------
    // Public properties
    //----------------------------------------------------------------------------------
    _Chart.defaultTranslate;
    _Chart.defaultScale;
    _Chart.zoom;
    _Chart.resetZoom;

    //----------------------------------------------------------------------------------
    // Local variables
    //----------------------------------------------------------------------------------
    let firstDraw = true;

    const svg = _Chart._layerGroup.append('svg')
      .attr({
        width: '100%',
        height: '100%'
      })
      .classed('map', true)
      .style({
        'overflow': 'hidden'
      });

    const map = initMap(svg);

    const zoom = _Chart.zoom = d3.behavior.zoom();
    addZoomScaleEvent(zoom);
    svg.call(zoom);

    setupMapScale(map, zoom);
    setupHighlight(map, zoom);
    setupDrill(map, zoom, svg);
    initZoomButtons(_Chart._layerGroup, zoom, svg);

    const tooltip = new DomoTooltip(svg.append('g'))
      .c('container', _Chart._layerGroup);

    const tooltipElems = addTooltipElems(tooltip);
    tooltipMapEvents(tooltip, tooltipElems, map);

    //----------------------------------------------------------------------------------
    // Triggers and event listeners
    //----------------------------------------------------------------------------------

    //**********************************************************************************
    // d3.Chart Transform
    //**********************************************************************************
    const superTransform = _Chart.transform;
    _Chart.transform = function (data) {
      const validData = superTransform(data);

      updateZoomBehavior(map, zoom);
      updateMapStyles(map);
      updateTooltipStyles(tooltip);

      const valueTree = createValueTree(map, data);
      mapData(map, valueTree);
      mapColor(map, valueTree);

      if (firstDraw) {
        _Chart.resetZoom();
      }
      else {
        zoom.event(svg);
      }

      firstDraw = false;
      return validData;
    };

    _Chart.resetZoom = function(animate) {
      zoom.event(svg);

      zoom.translate(_Chart.defaultTranslate)
      .scale(_Chart.defaultScale)
      .event(animate ? svg.transition().duration(750) : svg);
    };

    function createFsaTree() {
      const topology = require('json!./topologies/fsa.json');
      const geometries = topology.objects.good.geometries;
      const tree = new GeometryNode(topology);

      geometries.forEach(d => {
        const fsa = d.properties.fsa;
        const province = fsaToProvince(fsa);
        const digit = fsa[1];
        const region = fsa[2];

        tree.add(d, province, digit, region);
      });

      return tree;
    }

    function initMap(container) {

      const background = container.append('rect').classed('background', true)
        .attr({
          width: '100%',
          height: '100%'
        })
        .style('opacity', 0);
      const features = container.append('g').classed('features', true);

      const districtsG = features.append('g').classed('districts', true);
      const regionsG = features.append('g').classed('regions', true);
      const districtStrokeG = features.append('g').classed('district-stroke', true);
      const regionExtrasG = features.append('g').classed('region-extras', true);

      const scale = 9000;

      const projection = d3.geo.conicConformal()
        .rotate([96, -62])
        .parallels([-77, -49])
        .translate([0, 0])
        .scale(scale);

      const path = d3.geo.path()
        .projection(projection);

      const fsaTree = createFsaTree();

      const districts = initDistricts(districtsG, districtStrokeG, fsaTree, path);
      const regions = initRegions(regionsG, regionExtrasG, fsaTree, path);

      const fsaList = {};
      for (var i = 0; i < fsaTree.geometries.length; i++) {
        fsaList[fsaTree.geometries[i].properties.fsa] = true;
      }

      return {
        background,
        features,
        districts,
        regions,
        bounds: path.bounds(fsaTree.merge()),
        scale,
        path,
        fsaList
      };
    }

    function initDistricts(featureContainer, strokeContainer, fsaTree, path) {
      return {
        feature: featureContainer
          .selectAll('path')
          .data(fsaTree.mergeToArray())
          .enter().append('path')
          .attr('d', path)
          .attr('shape-rendering', 'crispEdges')
          .style({
            'cursor': 'pointer'
          }),
        stroke: strokeContainer
          .append('path')
          .attr('d', path(fsaTree.mesh(true)))
          .style({
            'fill': 'none',
            'pointer-events': 'none'
          }),
        strokeContainer: strokeContainer.node()
      };
    }

    function processRegionData(data, path) {
      const polygons = [];
      const points = [];

      data.forEach(d => {
        const length = Math.sqrt(path.area(d));

        d.path = {
          length,
          centroid: path.centroid(d)
        };

        const category = length === 0 ? points : polygons;
        category.push(d);
      });

      return {
        polygons,
        points
      };
    }

    function initRegions(regionContainer, regionExtrasContainer, fsaTree, path) {
      const regions = {};

      for (let a in fsaTree.children) {
        regions[a] = new LazyLoad()
          .get(region => {
            const data = processRegionData(fsaTree.children[a].mergeAllToArray(), path);

            const geometries = regionContainer.append('g')
              .call(hide);

            region.feature = geometries.append('g')
              .classed('feature', true)
              .selectAll('path')
              .data(data.polygons)
              .enter().append('path')
              .attr('d', path)
              .attr('shape-rendering', 'crispEdges');

            region.stroke = geometries.append('path')
              .classed('stroke', true)
              .attr('d', path(fsaTree.children[a].meshAll(true)))
              .style({
                'fill': 'none',
                'pointer-events': 'none'
              });

            region.digits = geometries.append('path')
              .classed('digits', true)
              .attr('d', path(fsaTree.children[a].mesh(true)))
              .style({
                'fill': 'none',
                'pointer-events': 'none'
              });

            const extras = regionExtrasContainer.append('g')
              .call(hide);

            region.point = extras.append('g')
              .classed('point', true)
              .selectAll('circle')
              .data(data.points)
              .enter().append('circle')
              .attr({
                cx: d => d.path.centroid[0],
                cy: d => d.path.centroid[1]
              });

            region.text = extras.append('g')
              .classed('text', true)
              .selectAll('text')
              .data(data.polygons)
              .enter().append('text')
              .attr({
                x: d => d.path.centroid[0],
                y: d => d.path.centroid[1]
              })
              .style({
                'pointer-events': 'none',
                'text-anchor': 'middle'
              })
              .text(d => d.properties.fsa);

            region.geometries = geometries.node();
            region.extras = extras.node();
            region.all = d3.selectAll([
              geometries.node(), extras.node()
            ]);
          });
        }

      return regions;
    }

    function updateZoomBehavior(map, zoom) {
      const maxZoom = 6e5;

      const bounds = map.bounds;
      const mapWidth = bounds[1][0] - bounds[0][0];
      const mapHeight = bounds[1][1] - bounds[0][1];

      const zoomedOut = Math.min(_Chart.c('height') / mapHeight, _Chart.c('width') / mapWidth);
      const zoomedIn = maxZoom / map.scale;

      const hw = _Chart.c('width') / 2;
      const hh = _Chart.c('height') / 2;

      zoom.on('zoom.bounds', () => {

        const t = zoom.translate();
        const s = zoom.scale();
        zoom.translate([
          clamp(t[0], -bounds[1][0] * s + hw, -bounds[0][0] * s + hw),
          clamp(t[1], -bounds[1][1] * s + hh, -bounds[0][1] * s + hh),
        ]);

        map.features
          .attr({
            'transform': 'translate(' + zoom.translate() + ')scale(' + zoom.scale() + ')'
          });
      });

      zoom.scaleExtent([zoomedOut, zoomedIn]);

      _Chart.defaultTranslate = [
        (-bounds[0][0] - mapWidth * 0.5) * zoomedOut + hw,
        (-bounds[0][1] - mapHeight * 0.5) * zoomedOut + hh
      ];

      _Chart.defaultScale = zoomedOut;
    }

    function updateMapStyles(map) {
      map.districts.stroke.style({
        'stroke': _Chart.c('districtStrokeColor')
      });

      for (let a in map.regions) {
        map.regions[a].get(region => {
          region.stroke.style({
            'stroke': _Chart.c('regionStrokeColor')
          });
          region.point.style({
            'stroke': _Chart.c('regionStrokeColor')
          });
          region.digits.style({
            'stroke': _Chart.c('digitStrokeColor')
          });
          region.text.style({
            'font-family': _Chart.c('fsaLabelFontFamily'),
            'fill': _Chart.c('fsaLabelFontColor'),
          });
          if (!_Chart.c('showFsaLabels').value) {
            region.text.style('display', 'none');
          }
        }, 'styles');
      }
    }

    function createValueTree(map, data) {
      const valueTree = new ValueNode();

      data.forEach(d => {
        const fsa = _Chart.a('FSA')(d).toUpperCase();
        const value = _Chart.a('Value')(d);

        const province = fsaToProvince(fsa);
        const region = fsa[1] + fsa[2];

        if (map.fsaList[fsa] && !valueTree.getDescendent(province, region)) {
          valueTree.add(value, province, region);
        }
      });

      valueTree.done();

      return valueTree;
    }

    function mapData(map, valueTree) {
      map.districts.feature.each(d => {
        d.properties.value = valueTree.getDescendent(d.properties.label).value;
      });

      for (let a in map.regions) {
        map.regions[a].get(region => {
          const setValue = d => {
            const fsa = d.properties.fsa;
            d.properties.value = valueTree.getDescendent(a, fsa[1] + fsa[2]).value;
          };

          region.feature.each(setValue);
          region.point.each(setValue);
        }, 'data');
      }
    }

    function mapColor(map, valueTree) {
      const colorScale = d3.scale.quantile()
        .domain([valueTree.min, valueTree.max])
        .range( _Chart.c('colorTheme').value);

      map.districts.feature.style('fill', d => {
        return d.properties.value ?
          colorScale(d.properties.value) :
          _Chart.c('noDataColor');
      });

      for (let a in map.regions) {
        map.regions[a].get(region => {

          const fill = d => {
            return d.properties.value ?
              colorScale(d.properties.value) :
              _Chart.c('noDataColor');
          };

          region.feature.style('fill', fill);
          region.point.style('fill', fill);
        }, 'colors');
      }
    }

    function fsaLabelSize(zoom) {
      const preferedSize = _Chart.c('fsaLabelFontSize');
      const minSize = 2 / 3 * preferedSize;
      const percentOfRegion = 0.15;
      const scale = zoom.scale();

      return function(d) {
        const actualSize = percentOfRegion * d.path.length;
        if (actualSize * scale < minSize) {
          this.style.display = 'none';
        }
        else {
          this.style.display = '';
          this.style.fontSize = Math.min(preferedSize / zoom.scale(), actualSize) + 'px';
        }
      };
    }

    function fsaPointSize(map, zoom) {
      const maxPointSize = 1e-4;
      const preferredPointSize = 18;

      return Math.min(maxPointSize * map.scale, preferredPointSize / zoom.scale());
    }

    function setupMapScale(map, zoom) {
      zoom.onScale('zoom.mapStroke', () => {
        map.districts.stroke.style('stroke-width', _Chart.c('districtStrokeWidth') / zoom.scale()  + 'px');
        for (let a in map.regions) {
          map.regions[a].get(region => {
            region.stroke.style('stroke-width', _Chart.c('regionStrokeWidth') / zoom.scale() + 'px');
            region.point.style('stroke-width', _Chart.c('regionStrokeWidth') / zoom.scale() + 'px');
            region.digits.style('stroke-width', _Chart.c('digitStrokeWidth') / zoom.scale() + 'px');
            region.point.attr('r', fsaPointSize(map, zoom));
          }, 'scaleStroke');
        }
      });
      zoom.onScale('zoom.mapLabels', () => {
        if (_Chart.c('showFsaLabels').value) {
          for (let a in map.regions) {
            map.regions[a].get(region => {
              region.text.each(fsaLabelSize(zoom));
            }, 'scaleLabels');
          }
        }
      });
    }

    function setupHighlight(map, zoom) {
      let highlight;

      const addHighlight = function(appendTo, minStrokeWidth) {
        d3.select(highlight).remove();

        if (_Chart.c('highlightStrokeWidth') === 0) {
          return;
        }

        highlight = this.cloneNode();

        const strokeWidth = Math.max(minStrokeWidth(), _Chart.c('highlightStrokeWidth'));

        d3.select(highlight)
          .style({
            'fill': 'none',
            'stroke': _Chart.c('highlightStrokeColor'),
            'stroke-width': strokeWidth / zoom.scale() + 'px',
            'shape-rendering': 'geometricPrecision'
          });

        appendTo.appendChild(highlight);
      };

      const highlightable = function(selection, appendTo, minStrokeWidth) {
        selection.on('mouseover', function() {
          addHighlight.call(this, appendTo, minStrokeWidth);
          zoom.onScale('zoom.highlight', () => {
            addHighlight.call(this, appendTo, minStrokeWidth);
          });
        });

        selection.on('mouseleave', function() {
          d3.select(highlight).remove();
          zoom.onScale('zoom.highlight', null);
          highlight = null;
        });
      };

      highlightable(map.districts.feature, map.districts.strokeContainer, () => _Chart.c('districtStrokeWidth'));
      for (let a in map.regions) {
        map.regions[a].get(region => {
          highlightable(region.feature, map.districts.strokeContainer, () => _Chart.c('regionStrokeWidth'));
          highlightable(region.point, region.extras, () => _Chart.c('regionStrokeWidth'));
        });
      }
    }

    function setupDrill(map, zoom, zoomContainer) {
      let lastDistrict;

      const clearLastDistrict = function(currentDistrict) {
        if (lastDistrict) {

          map.regions[lastDistrict].get(region => {
            region.all.call(hide);
          }, 'visibility');
        }

        lastDistrict = currentDistrict;
      };

      clickWithoutMoving(map.background, function() {
        clearLastDistrict();
        _Chart.resetZoom(true);
      });

      clickWithoutMoving(map.districts.feature, function(d) {
        clearLastDistrict(d.properties.label);

        map.regions[d.properties.label]
          .load()
          .get(region => {
            region.all.call(show);
          }, 'visibility');

        zoomToBounds(zoom, zoomContainer, map.path.bounds(d));
      });
    }

    function addTooltipElems(tooltip) {
      tooltip.c({
        format: 'textRectText'
      })
      .draw();

      const value = tooltip._appends.append('text');
      const rect = tooltip._appends.append('rect');
      const label = tooltip._appends.append('text');

      return {
        value, rect, label
      };
    }

    function updateTooltipStyles(tooltip) {
      tooltip.c({
        tooltipBackgroundColor: _Chart.c('tooltipBackground')
      });

      tooltip._appends.selectAll('text')
        .style({
          'font-family': _Chart.c('tooltipFontFamily'),
          'font-size': _Chart.c('tooltipFontSize') + 'px',
          'fill': _Chart.c('tooltipFontColor')
        });

      tooltip._appends.selectAll('rect')
        .style('fill', _Chart.c('tooltipFontColor'));
    }

    function tooltipMapEvents(tooltip, tooltipElems, map) {

      const tooltipHover = function(sel, mouseoverCallback) {
        sel.on('mouseover.tooltip', function(d) {
          mouseoverCallback.call(this, d);
          tooltip.trigger('draw');
        });
        sel.on('mousemove.tooltip', function() {
          const m = d3.mouse(_Chart._layerGroup.node());
          tooltip.trigger('moveTo', {x: m[0], y: m[1]});
        });
        sel.on('mouseleave.tooltip', function() {
          tooltip.trigger('remove');
        });
      };

      tooltipHover(map.districts.feature, d => {
        const label = d.properties.label;
        const value = d.properties.value;

        tooltipElems.label.text(label);
        tooltipElems.value.text(value ? service.summaryNumber(value) : 'No Data');
      });

      for (let a in map.regions) {
        map.regions[a].get(region => {
          const hover = d => {
            const fsa = d.properties.fsa;
            const value = d.properties.value;

            tooltipElems.label.text(fsa);
            tooltipElems.value.text(value ? service.summaryNumber(value) : 'No Data');
          };

          tooltipHover(region.feature, hover);
          tooltipHover(region.point, hover);
        }, 'tooltip');
      }
    }

    function initZoomButtons(buttonContainer, zoom, zoomContainer) {
      const zoomButtonAmount = 1.8;

      const buttonDiv = buttonContainer.append('div')
        .style({
          'position': 'absolute',
          'bottom': '10px',
          'right': '10px'
        });

      const buttons = {
        zoomIn: button.create(buttonDiv, plusSvg).call(button.top),
        zoomOut: button.create(buttonDiv, minusSvg).call(button.bottom)
      };

      const zoomEvent = function(amount) {

        return () => {
          const x = _Chart.c('width') / 2;
          const y = _Chart.c('height') / 2;
          const t = zoom.translate();
          const k = zoom.scale();
          const e = zoom.scaleExtent();
          const newScale = clamp(zoom.scale() * amount, e[0], e[1]);

          /*
            Translate so that the x and y positions on the screen correspond
            to the same points on the map before and after the zoom.
          */
          const newTranslate = [
            x + (t[0] - x) * newScale / k,
            y + (t[1] - y) * newScale / k
          ];

          zoom.event(zoomContainer);

          zoom.scale(newScale)
            .translate(newTranslate)
            .size([2 * x, 2 * y]);

          zoom.event(zoomContainer.transition());
        };
      };

      buttons.zoomIn.on('click', zoomEvent(zoomButtonAmount));
      buttons.zoomOut.on('click', zoomEvent(1 / zoomButtonAmount));
    }

    function zoomToBounds(zoom, zoomContainer, bounds) {
      const dx = bounds[1][0] - bounds[0][0];
      const dy = bounds[1][1] - bounds[0][1];
      const x = (bounds[0][0] + bounds[1][0]) / 2;
      const y = (bounds[0][1] + bounds[1][1]) / 2;
      const e = zoom.scaleExtent();

      const scale = clamp(0.95 * Math.min(_Chart.c('width') / dx, _Chart.c('height') / dy), e[0], e[1]);
      const translate = [_Chart.c('width') / 2 - scale * x, _Chart.c('height') / 2 - scale * y];

      zoom.event(zoomContainer);

      zoom.translate(translate)
        .scale(scale)
        .event(zoomContainer.transition().duration(750));
    }

    function hide(sel) {
      return sel
        .style('pointer-events', 'none')
        .transition('hide')
        .duration(_Chart.c('animationDuration'))
        .style('opacity', 0)
        .each('end', () => {
          sel.style('display', 'none');
        });
    }

    function show(sel) {
      return sel
        .style({
          'pointer-events': '',
          'display': ''
        })
        .transition('hide')
        .duration(_Chart.c('animationDuration'))
        .style('opacity', 1);
    }

    function clickWithoutMoving(selection, fn) {
      let lastEvent;
      selection.on('mousedown', function() {
        lastEvent = d3.event;
      });

      selection.on('mouseup', function() {
        if (lastEvent.pageX === d3.event.pageX && lastEvent.pageY === d3.event.pageY) {
          fn.apply(this, arguments);
        }
      });
    }

    function addZoomScaleEvent(zoom) {
      const dispatch = d3.dispatch('zoom');

      let lastScale;
      let timeout;

      const onZoom = () => {
        const s = zoom.scale();

        clearTimeout(timeout);

        if (lastScale !== s) {
          dispatch.zoom();
        }
        else {
          timeout = setTimeout(dispatch.zoom, 150);
        }

        lastScale = s;
      };

      zoom.on('zoom.scaleEvent', onZoom);
      zoom.onScale = dispatch.on.bind(dispatch);
    }

    function clamp(n, min, max) {
      return Math.max(Math.min(n, max), min);
    }
  }
});
